__author__ = 'alexisgallepe'

from . import Piece
import bitstring
import logging
from threading import Thread
from pubsub import pub


class PiecesManager(Thread):
    def __init__(self, torrent):
        Thread.__init__(self)
        self.torrent = torrent
        self.piecesCompleted = False

        self.numberOfPieces = torrent.numberOfPieces
        self.bitfield = bitstring.BitArray(self.numberOfPieces)
        self.pieces = self.generate_pieces()
        self.files = self.get_files()

        for file in self.files:
            idPiece = file['idPiece']
            self.pieces[idPiece].files.append(file)

        # Create events
        pub.subscribe(self.receive_block_piece, 'PiecesManager.Piece')
        pub.subscribe(self.update_bit_field, 'PiecesManager.PieceCompleted')


    def check_disk_pieces(self):
        for piece in self.pieces:
            piece.isCompleteOnDisk() # this should set all the finished bools on the finished pieces

    def update_bit_field(self,pieceIndex):
        self.bitfield[pieceIndex] = 1

    def receive_block_piece(self,piece):
        piece_index, piece_offset, piece_data = piece
        self.pieces[int(piece_index)].setBlock(piece_offset, piece_data)

    def generate_pieces(self):
        pieces = []
        for i in range(self.numberOfPieces):
            start = i * 20
            end = start + 20

            if i == (self.numberOfPieces-1):
                pieceLength = self.torrent.totalLength - (self.numberOfPieces-1) * self.torrent.pieceLength
                pieces.append(Piece.Piece(i, pieceLength, self.torrent.pieces[start:end]))
            else:
                pieces.append(Piece.Piece(i, self.torrent.pieceLength, self.torrent.pieces[start:end]))
        return pieces

    def are_pieces_completed(self):
        for piece in self.pieces:
            if not piece.finished:
                return False

        self.piecesCompleted = True
        logging.info("File(s) downloaded")
        return True

    def get_files(self):
        files = []
        pieceOffset = 0
        pieceSizeUsed = 0

        for f in self.torrent.fileNames:

            currentSizeFile = f["length"]
            fileOffset = 0

            while currentSizeFile > 0:
                idPiece = int(pieceOffset / self.torrent.pieceLength)
                pieceSize = self.pieces[idPiece].pieceSize - pieceSizeUsed

                if currentSizeFile - pieceSize < 0:
                    file = {"length": currentSizeFile,"idPiece":idPiece ,"fileOffset":fileOffset, "pieceOffset":pieceSizeUsed ,"path":f["path"]}
                    pieceOffset +=  currentSizeFile
                    fileOffset +=  currentSizeFile
                    pieceSizeUsed += currentSizeFile
                    currentSizeFile = 0

                else:
                    currentSizeFile -= pieceSize
                    file = {"length":pieceSize,"idPiece":idPiece ,"fileOffset":fileOffset,"pieceOffset":pieceSizeUsed , "path":f["path"]}
                    pieceOffset += pieceSize
                    fileOffset += pieceSize
                    pieceSizeUsed = 0

                files.append(file)
        return files


    def get_block(self, piece_index, block_offset, block_length):
        for piece in self.pieces:
            if piece_index == piece.pieceIndex:
                if piece.finished:
                    return piece.get_block(block_offset,block_length)
                else:
                    break

        return None



__author__ = 'alexisgallepe'

import socket
import struct
import bitstring
from bitstring import BitArray
from pubsub import pub
from . import utils
import threading
import logging
from six import string_types


class Peer(object):
    def __init__(self, torrent, ip, port=6881):
        self.lastHandshakeAttempt = 0
        self.lock = threading.Lock()
        self.handshake = None
        self.hasHandshaked = False
        self.readBuffer = b""
        self.counter = 10
        self.socket = None
        self.ip = ip
        self.port = port
        self.torrent = torrent
        self.socketsPeers = []

        self.state = {
            'am_choking': True,
            'am_interested': False,
            'peer_choking': True,
            'peer_interested': False,
        }
        self.idFunction = {
            0: self.choke,
            1: self.unchoke,
            2: self.interested,
            3: self.not_interested,
            4: self.have,
            5: self.bitfield,
            6: self.request,
            7: self.piece,
            8: self.cancel,
            9: self.portRequest
        }

        self.numberOfPieces = torrent.numberOfPieces

        self.bitField = bitstring.BitArray(self.numberOfPieces)

    def connectToPeer(self, timeout=10):
        try:
            self.socket = socket.create_connection((self.ip, self.port), timeout)
            logging.info("connected to peer ip: {0} - port: {1}".format(self.ip, self.port))
            self.build_handshake()
            return True
        except Exception:
            print("connectToPeer Socket Timeout Error")

        return False

    def hasPiece(self, index):
        return self.bitField[index]

    def build_handshake(self):
        pstr = "BitTorrent protocol".encode('utf-8')
        reserved = "0" * 8
        hs = struct.pack("B" + str(len(pstr)) + "s8x20s20s",
                         len(pstr),
                         pstr,
                         # reserved,
                         self.torrent.info_hash,
                         self.torrent.peer_id
                         )
        assert len(hs) == 49 + len(pstr)
        self.handshake = hs

    def build_interested(self):
        return struct.pack('!I', 1) + struct.pack('!B', 2)

    def build_request(self, index, offset, length):
        header = struct.pack('>I', 13)
        id = b'\x06'

        if isinstance(length, (bytes, bytearray)):
            id = '\x06'

        index = struct.pack('>I', index)
        offset = struct.pack('>I', offset)
        length = struct.pack('>I', length)
        request = header + id + index + offset + length

        return request

    def build_piece(self, index, offset, data):
        header = struct.pack('>I', 13)
        id = '\x07'
        index = struct.pack('>I', index)
        offset = struct.pack('>I', offset)
        data = struct.pack('>I', data)
        piece = header + id + index + offset + data

        return piece

    def build_bitfield(self):
        length = struct.pack('>I', 4)
        id = '\x05'
        bitfield = self.bitField.tobytes()
        bitfield = length + id + bitfield
        return bitfield

    def sendToPeer(self, msg):
        try:
            self.socket.send(msg)
        except Exception as e:
            print("sendToPeer Error: ")
            print(self.ip)
            print(e)

    def checkHandshake(self, buf, pstr="BitTorrent protocol"):
        if isinstance(buf, (bytes, bytearray)):
            pstr_rec = buf[1:20].decode('utf-8')
        else:
            pstr_rec = buf[1:20]

        if pstr_rec == pstr:
            handshake = buf[:68]
            expected_length, info_dict, info_hash, peer_id = struct.unpack(
                "B" + str(len(pstr)) + "s8x20s20s",
                handshake)

            if self.torrent.info_hash == info_hash:
                self.hasHandshaked = True
                # self.sendToPeer(self.build_bitfield())
            else:
                logging.warning("Error with peer's handshake")

            self.readBuffer = self.readBuffer[28 + len(info_hash)+20:]

    def keep_alive(self, payload):
        try:
            keep_alive = struct.unpack("!I", payload[:4])[0]
            if keep_alive == 0:
                logging.info('KEEP ALIVE')
                return True
        except Exception:
            pass

        return False

    def choke(self, payload=None):
        logging.info('choke')
        print("choking peer: " + self.ip)
        self.state['peer_choking'] = True

    def unchoke(self, payload=None):
        logging.info('unchoke')
        print("Unchoking peer: " + self.ip)
        pub.sendMessage('PeersManager.peerUnchoked', peer=self)
        self.state['peer_choking'] = False

    def interested(self, payload=None):
        logging.info('interested')
        self.state['peer_interested'] = True

    def not_interested(self, payload=None):
        logging.info('not_interested')
        self.state['peer_interested'] = False

    def have(self, payload):
        index = utils.convertBytesToDecimal(payload)
        self.bitField[index] = True
        pub.sendMessage('RarestPiece.updatePeersBitfield', bitfield=self.bitField, peer=self)

    def bitfield(self, payload):
        self.bitField = BitArray(bytes=payload)
        logging.info('request')
        pub.sendMessage('RarestPiece.updatePeersBitfield', bitfield=self.bitField, peer=self)

    def request(self, payload):
        piece_index = payload[:4]
        block_offset = payload[4:8]
        block_length = payload[8:]
        logging.info('request')
        pub.sendMessage('PiecesManager.PeerRequestsPiece', piece=(piece_index, block_offset, block_length), peer=self)

    def piece(self, payload):
        piece_index = utils.convertBytesToDecimal(payload[:4])
        piece_offset = utils.convertBytesToDecimal(payload[4:8])
        piece_data = payload[8:]
        pub.sendMessage('PiecesManager.Piece', piece=(piece_index, piece_offset, piece_data))

    def cancel(self, payload=None):
        logging.info('cancel')

    def portRequest(self, payload=None):
        logging.info('portRequest')




__author__ = 'alexisgallepe'

import time
import logging
import os
import requests
import json
import datetime
from queue import Queue
from . import PeersManager
from . import PeerSeeker
from . import PiecesManager
from . import Torrent
from . import Tracker
from . import HttpPeer
from . import utils


class Client(object):
    @classmethod
    def __init__(self, hash, torrent_dir):
        newpeersQueue = Queue()
        self.hash = hash
        self.torrent_dir = torrent_dir

        self.torrent = Torrent.Torrent(self.hash, self.torrent_dir)
        self.tracker = Tracker.Tracker(self.torrent, newpeersQueue)
        self.piecesManager = PiecesManager.PiecesManager(self.torrent)
        self.peerSeeker = PeerSeeker.PeerSeeker(newpeersQueue, self.torrent)
        self.peersManager = PeersManager.PeersManager(self.torrent, self.piecesManager)

        self.peersManager.start()
        logging.info("Peers-manager Started")

        self.peerSeeker.start()
        logging.info("Peer-seeker Started")

        self.piecesManager.start()
        logging.info("Pieces-manager Started")

        self.piecesManager.check_disk_pieces()

    def start(self):
        starting_size = self.check_percent_finished()
        new_size = starting_size
        old_size = 0
        while not self.piecesManager.are_pieces_completed():
            if len(self.peersManager.unchokedPeers) > 0:
                for piece in self.piecesManager.pieces:
                    if not piece.finished:
                        pieceIndex = piece.pieceIndex

                        peer = self.peersManager.getUnchokedPeer(pieceIndex)
                        if not peer:
                            continue

                        data = self.piecesManager.pieces[pieceIndex].getEmptyBlock()
                        if data:
                            index, offset, length = data
                            self.peersManager.requestNewPiece(peer, index, offset, length)

                        piece.isComplete()
                        self.reset_pending_blocks(piece)
            if len(self.peersManager.httpPeers) > 0:
                for httpPeer in self.peersManager.httpPeers:
                    pieces = httpPeer.get_pieces(self.piecesManager)
                    pieces_by_file = httpPeer.construct_pieces_by_file(pieces)  # set all those blocks to Pending
                    responses = httpPeer.request_ranges(pieces_by_file)
                    httpPeer.publish_responses(responses, pieces_by_file)

            new_size = self.check_percent_finished()
            if new_size == old_size:
                continue

            old_size = new_size
            print("# Peers:", len(self.peersManager.unchokedPeers), " # HTTPSeeds:", len(self.peersManager.httpPeers), " Completed: ", float((float(new_size) / self.torrent.totalLength)*100), "%")

            time.sleep(0.1)
        downloaded = new_size - starting_size
        remaining = self.torrent.totalLength - (starting_size + downloaded)
        self.tracker.stop_message(downloaded, remaining)
        self.peerSeeker.requestStop()
        self.peersManager.requestStop()

        if remaining == 0:
            utils.write_timestamp(self.hash)

        return self.torrent_dir + self.torrent.torrentFile['info']['name']

    def reset_pending_blocks(self, piece):
        for block in piece.blocks:
            if(int(time.time()) - block[3]) > 8 and block[0] == "Pending":
                block[0] = "Free"
                block[3] = 0

    def check_percent_finished(self):
        b = 0
        for i in range(self.piecesManager.numberOfPieces):
            for j in range(self.piecesManager.pieces[i].num_blocks):
                if self.piecesManager.pieces[i].blocks[j][0] == "Full":
                    b += len(self.piecesManager.pieces[i].blocks[j][2])
        return b



__author__ = 'alexisgallepe'

import math
import time
import logging

from . import utils
from pubsub import pub

BLOCK_SIZE = 2 ** 14


class Piece(object):
    def __init__(self, pieceIndex, pieceSize, pieceHash):
        self.pieceIndex = pieceIndex
        self.pieceSize = pieceSize
        self.pieceHash = pieceHash
        self.finished = False
        self.files = []
        self.pieceData = b""
        self.BLOCK_SIZE = BLOCK_SIZE
        self.num_blocks = int(math.ceil(float(pieceSize) / BLOCK_SIZE))
        self.blocks = []
        self.initBlocks()

    def initBlocks(self):
        self.blocks = []
        if self.num_blocks > 1:
            for i in range(self.num_blocks):
                    self.blocks.append(["Free", BLOCK_SIZE, b"", 0])

            # Last block of last piece, the special block
            if (self.pieceSize % BLOCK_SIZE) > 0:
                self.blocks[self.num_blocks-1][1] = self.pieceSize % BLOCK_SIZE
        else:
            self.blocks.append(["Free", int(self.pieceSize), b"", 0])

    def get_file_offset(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return f.get('fileOffset')

    def get_file_length(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return f.get('length')

    def get_piece_offset(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return f.get('pieceOffset')

    def setBlock(self, offset, data, write=True):
        if not self.finished:
            if offset == 0:
                index = 0
            else:
                index = int(offset / BLOCK_SIZE)

            self.blocks[index][2] = data
            self.blocks[index][0] = "Full"
            self.isComplete(write=write)

    def get_block(self, block_offset, block_length):
        return self.pieceData[block_offset:block_length]

    def getEmptyBlock(self):
        if not self.finished:
            blockIndex = 0
            for block in self.blocks:
                if block[0] == "Free":
                    block[0] = "Pending"
                    block[3] = int(time.time())
                    return self.pieceIndex, blockIndex * BLOCK_SIZE, block[1]
                blockIndex += 1
        return False

    def freeBlockLeft(self):
        for block in self.blocks:
            if block[0] == "Free":
                return True
        return False

    def isCompleteOnDisk(self):
        block_offset = 0
        data = b''
        for f in self.files:
            try:
                f_ptr = open(f["path"], 'rb')
            except IOError:
                all_files_finished = False
                break
            f_ptr.seek(f["fileOffset"])
            data += f_ptr.read(f["length"])
            f_ptr.close()
            block_offset += f['length']
        if self.isHashPieceCorrect(data):
            for block in range(self.num_blocks):
                start_offset = block*BLOCK_SIZE
                end_offset = block*BLOCK_SIZE + BLOCK_SIZE
                self.setBlock(offset=start_offset, data=data[start_offset: end_offset], write=False)

    def isComplete(self, write=True):
        # If there is at least one block Free|Pending -> Piece not complete -> return false
        for block in self.blocks:
            if block[0] == "Free" or block[0] == "Pending":
                return False
        # Before returning True, we must check if hashes match
        data = self.assembleData()
        if self.isHashPieceCorrect(data):
            self.finished = True
            self.pieceData = data
            if write:
                self.writeFilesOnDisk()
            pub.sendMessage('PiecesManager.PieceCompleted', pieceIndex=self.pieceIndex)
            return True

        else:
            return False

    def writeFunction(self, pathFile, data, offset):
        try:
            f = open(pathFile, 'r+b')
        except IOError:
            f = open(pathFile, 'wb')
        f.seek(offset)
        f.write(data)
        f.close()

    def writeFilesOnDisk(self):
        for f in self.files:
            pathFile = f["path"]
            fileOffset = f["fileOffset"]
            pieceOffset = f["pieceOffset"]
            length = f["length"]
            self.writeFunction(pathFile, self.pieceData[pieceOffset: pieceOffset + length], fileOffset)

    def assembleData(self):
        buf = b""
        for block in self.blocks:
            buf += block[2]
        return buf

    def isHashPieceCorrect(self, data):
        if utils.sha1_hash(data) == self.pieceHash:
            return True
        else:
            self.initBlocks()
            return False



__author__ = 'alexisgallepe'

import time
import bencode
import logging
import os
from . import utils
try:
    from urllib.parse import urlparse, urlencode
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError


class Torrent(object):
    def __init__(self, hash, torrent_dir):
        self.hash = hash
        self.torrent_dir = torrent_dir
        if not self.get_from_url() and not self.get_from_file():
            raise Exception("Could not find a torrent with this hash on the tracker or in the torrent directory:" + str(self.torrent_dir))

        with open(torrent_dir + hash + '.torrent', 'rb') as file:
            contents = file.read()

        self.torrentFile = bencode.decode(contents)
        self.totalLength = 0
        self.pieceLength = self.torrentFile['info']['piece length']
        self.pieces = self.torrentFile['info']['pieces']

        self.info_hash = utils.sha1_hash(bencode.encode(self.torrentFile['info']))
        self.peer_id = self.generate_peer_id()
        self.announceList = self.getTrakers()
        self.fileNames = []

        self.get_files()
        if self.totalLength % self.pieceLength == 0:
            self.numberOfPieces = self.totalLength / self.pieceLength
        else:
            self.numberOfPieces = int(self.totalLength / self.pieceLength) + 1

        logging.debug(self.announceList)
        logging.debug(self.fileNames)

        assert(self.totalLength > 0)
        assert(len(self.fileNames) > 0)

    def get_from_file(self):
        try:
            torrent_path = os.path.join(self.torrent_dir, self.hash + '.torrent')
            return bencode.decode(open(torrent_path, 'rb').read())
        except Exception:
            return False

    def get_from_url(self):
        contents = None
        try:
            url = "http://academictorrents.com/download/" + self.hash
            torrent_path = os.path.join(self.torrent_dir, self.hash + '.torrent')
            if not os.path.isdir(self.torrent_dir):
                os.makedirs(self.torrent_dir)
            response = urlopen(url).read()
            open(torrent_path, 'wb').write(response)
            contents = bencode.decode(open(torrent_path, 'rb').read())
        except Exception as e:
            print("could not download the torrent")
        return contents

    def get_files(self):
        root = self.torrent_dir + self.torrentFile['info']['name'] #+ "/"
        if 'files' in self.torrentFile['info']:
            if not os.path.exists(root):
                os.mkdir(root, 0o766)

            for f in self.torrentFile['info']['files']:
                pathFile = os.path.join(root, *f["path"])

                if not os.path.exists(os.path.dirname(pathFile)):
                    os.makedirs(os.path.dirname(pathFile))

                self.fileNames.append({"path": pathFile , "length": f["length"]})
                self.totalLength += f["length"]

        else:
            self.fileNames.append({"path": root , "length": self.torrentFile['info']['length']})
            self.totalLength = self.torrentFile['info']['length']

    def getTrakers(self):
        if 'announce-list' in self.torrentFile:
            return self.torrentFile['announce-list']
        else:
            return [[ self.torrentFile['announce'] ]]

    def generate_peer_id(self):
        seed = str(time.time())
        return "AT-" + utils.sha1_hash(seed.encode())



__author__ = 'martin weiss, alexis gallepe'

import hashlib
import os
import json
import datetime


def convertBytesToDecimal(headerBytes):
    size = 0
    power = len(headerBytes) - 1
    for ch in headerBytes:
        if isinstance(ch, int):
            size += ch * 256 ** power
        else:
            size += int(ord(ch)) * 256 ** power
        power -= 1
    return size


def sha1_hash(string):
    """Return 20-byte sha1 hash of string."""
    return hashlib.sha1(string).digest()


def get_timestamp_filename():
    return os.getcwd() + "/torrent_timestamps.json"


def get_torrent_dir(datastore=None, name=None):
    if not datastore:
        datastore = os.getcwd() + "/datastore/"
    elif datastore == "." or datastore == "./":
        datastore = "./"
    elif datastore[-1] != "/":
        datastore = datastore + "/"
    elif datastore[0] != "/":
        datastore = os.getcwd() + "/" + datastore
    if not name:
        return datastore
    if isinstance(name, str):
        name = name.strip("/")
    return datastore + name + '/'


def write_timestamp(hash):
    filename = get_timestamp_filename()
    with open(filename, 'w+') as f:
        try:
            timestamps = json.loads(f.read())
        except ValueError:
            timestamps = {}
        timestamps[hash] = int(datetime.datetime.now().strftime("%s"))
        json.dump(timestamps, f)

def read_timestamp(hash):
    filename = get_timestamp_filename()
    with open(filename) as f:
        return json.load(f).get('hash', 0)

def timestamp_is_recent(timestamp):
    seconds_in_a_month = 86400 * 30
    if timestamp > int(datetime.datetime.now().strftime("%s")) - seconds_in_a_month:
        return True
    return False

def filenames_present(torrent, datastore):
    name = torrent['info']['name']
    return name in os.listdir(datastore)



__author__ = 'alexisgallepe'

import select
import struct
from threading import Thread, Timer
from . import utils
from pubsub import pub
from . import RarestPieces
import logging
from . import HttpPeer, Peer
import time


class PeersManager(Thread):
    def __init__(self, torrent, piecesManager):
        Thread.__init__(self)
        self.peers = []
        self.unchokedPeers = []
        self.httpPeers = []
        self.torrent = torrent
        self.piecesManager = piecesManager
        self.rarestPieces = RarestPieces.RarestPieces(piecesManager)
        self.stopRequested = False
        self.setDaemon(True)

        self.piecesByPeer = []
        for i in range(self.piecesManager.numberOfPieces):
            self.piecesByPeer.append([0, []])

        for url in self.torrent.torrentFile.get('url-list'):
            try:
                peer = HttpPeer.HttpPeer(torrent, url)
                self.httpPeers.append(peer)
                print("adding HttpPeer: " + peer.url)
            except Exception:
                pass

        # Events
        pub.subscribe(self.addPeer, 'PeersManager.newPeer')
        pub.subscribe(self.addUnchokedPeer, 'PeersManager.peerUnchoked')
        pub.subscribe(self.handlePeerRequests, 'PeersManager.PeerRequestsPiece')
        pub.subscribe(self.peersBitfield, 'PeersManager.updatePeersBitfield')
        pub.subscribe(self.chokePeer, 'PeersManager.chokePeer')

    def requestStop(self):
        self.stopRequested = True

    def peersBitfield(self, bitfield=None, peer=None, pieceIndex=None):
        if pieceIndex is not None:
            self.piecesByPeer[pieceIndex] = ["", []]
            return

        for i in range(len(self.piecesByPeer)):
            if bitfield[i] == 1 and peer not in self.piecesByPeer[i][1] and not self.piecesByPeer[i][0] == "":
                self.piecesByPeer[i][1].append(peer)
                self.piecesByPeer[i][0] = len(self.piecesByPeer[i][1])

    def getUnchokedPeer(self, index):
        for peer in self.unchokedPeers:
            if isinstance(peer, HttpPeer.HttpPeer) or (isinstance(peer, Peer.Peer) and peer.hasPiece(index)):
                return peer
        return False

    def run(self):
        while not self.stopRequested:
            self.startConnectionToPeers()
            read = [p.socket for p in self.peers]
            readList, _, _ = select.select(read, [], [], 1)

            # Receive from peers
            for socket in readList:
                peer = self.getPeerBySocket(socket)
                try:
                    msg = socket.recv(1024)
                except Exception as e:
                    print("peersManager socket.recv error: ")
                    print(e)
                    print("removing peer: " + peer.ip)
                    self.removePeer(peer)
                    continue

                if len(msg) == 0:
                    print("length of message received in peersmanager was 0 -- removing peer: " + peer.ip)
                    self.removePeer(peer)
                    continue

                peer.readBuffer += msg
                self.manageMessageReceived(peer)

    def startConnectionToPeers(self):
        for peer in self.peers:
            if not peer.hasHandshaked:
                if peer.lastHandshakeAttempt < time.time() - 10:
                    peer.lastHandshakeAttempt = time.time()
                    try:
                        peer.sendToPeer(peer.handshake)
                        interested = peer.build_interested()
                        peer.sendToPeer(interested)
                    except Exception as e:
                        print("startConnectToPeers Error: ")
                        print(e)
                        print("removing peer: " + peer.ip)
                        self.removePeer(peer)

    def addPeer(self, peer):
        self.peers.append(peer)

    def addUnchokedPeer(self, peer):
        self.unchokedPeers.append(peer)

    def chokePeer(self, peer):
        self.unchokedPeers = [x for x in self.unchokedPeers if not hasattr(x, 'url') or (hasattr(x, 'url') and x.url != peer.url)]
        Timer(.005, self.addUnchokedPeer, [peer]).start()

    def removePeer(self, peer):
        if peer in self.peers:
            try:
                peer.socket.close()
            except Exception:
                pass

            self.peers.remove(peer)

        if peer in self.unchokedPeers:
            self.unchokedPeers.remove(peer)

        for rarestPiece in self.rarestPieces.rarestPieces:
            if peer in rarestPiece["peers"]:
                rarestPiece["peers"].remove(peer)

    def getPeerBySocket(self, socket):
        for peer in self.peers:
            if socket == peer.socket:
                return peer

        raise ("peer not present in PeerList")

    def handlePeerRequests(self, piece, peer):
        piece_index, block_offset, block_length = piece
        block = self.piecesManager.get_block(piece_index, block_offset, block_length)
        piece = peer.build_request(self, piece_index, block_offset, block)
        peer.sendToPeer(piece)

    def manageMessageReceived(self, peer):
        while len(peer.readBuffer) > 0:
            if peer.hasHandshaked is False:
                peer.checkHandshake(peer.readBuffer)
                return

            msgLength = utils.convertBytesToDecimal(peer.readBuffer[0:4])

            # handle keep alive
            if peer.keep_alive(peer.readBuffer):
                return

            # len 0
            try:
                msgCode = int(ord(peer.readBuffer[4:5]))
                payload = peer.readBuffer[5:4 + msgLength]
            except Exception as e:
                logging.info(e)
                return

            # Message is not complete. Return
            if len(payload) < msgLength - 1:
                return

            peer.readBuffer = peer.readBuffer[msgLength + 4:]
            try:
                peer.idFunction[msgCode](payload)
            except Exception as e:
                logging.debug("error id:", msgCode, " ->", e)
                return

    def requestNewPiece(self, peer, pieceIndex, blockOffset, length):
        request = peer.build_request(pieceIndex, blockOffset, length)
        peer.sendToPeer(request)



__author__ = 'alexisgallepe'

import bencode
import requests
import logging
import struct, random, socket
# Python 2 and 3: alternative 4
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import threading, time
import pkg_resources


class FuncThread(threading.Thread):
    def __init__(self, target, *args):
        self._target = target
        self._args = args
        threading.Thread.__init__(self)

    def run(self):
        self._target(*self._args)

class Tracker(object):
    def __init__(self, torrent, newpeersQueue):
        self.torrent = torrent
        self.lstThreads =  []
        self.newpeersQueue = newpeersQueue
        self.getPeersFromTrackers()

    def getPeersFromTrackers(self):
        for tracker in self.torrent.announceList:
            if tracker[0] == '':
                continue
            elif tracker[0][:4] == "http":
                t1 = FuncThread(self.scrapeHTTP, self.torrent, tracker[0])
                self.lstThreads.append(t1)
                t1.start()
            else:
                #self.scrape_udp(self.torrent, tracker[0])
                t2 = FuncThread(self.scrape_udp, self.torrent, tracker[0])
                self.lstThreads.append(t2)
                t2.start()

        for t in self.lstThreads:
            t.join()


    def scrapeHTTP(self, torrent, tracker):
        params = {
            'info_hash': torrent.info_hash,
            'peer_id': torrent.peer_id,
            'uploaded': 0,
            'downloaded': 0,
            'left': torrent.totalLength,
            'event': 'started',
            'port': 6881
        }
        try:
            answerTracker = requests.get(tracker, params=params, timeout=20)
            lstPeers = bencode.decode(answerTracker.content)
            for peer in lstPeers['peers']:
                self.newpeersQueue.put([peer['ip'], peer['port']])
        except:
            pass

    def stop_message(self, downloaded, remaining):
        if downloaded == 0:
            return True
        for tracker in self.torrent.announceList:
            if tracker[0] == '':
                continue
            elif tracker[0][:4] == "http":
                event = "finished" if remaining == 0 else "stopped"
                version = pkg_resources.get_distribution("academictorrents").version
                params = {
                    'info_hash': self.torrent.info_hash,
                    'peer_id': self.torrent.peer_id,
                    'uploaded': 0,
                    'downloaded': downloaded,
                    'left': remaining,
                    'event': event,
                    'port': 6881
                }
                try:

                    resp = requests.get(tracker[0], params=params, timeout=20, headers={'user-agent': "AT-Client/" + version + " " + requests.utils.default_user_agent()})
                except:
                    pass
            return params, resp


    def make_connection_id_request(self):
        conn_id = struct.pack('>Q', 0x41727101980)
        action = struct.pack('>I', 0)
        trans_id = struct.pack('>I', random.randint(0, 100000))

        return (conn_id + action + trans_id, trans_id, action)

    def make_announce_input(self,info_hash, conn_id, peer_id):
        action = struct.pack('>I', 1)
        trans_id = struct.pack('>I', random.randint(0, 100000))

        downloaded = struct.pack('>Q', 0)
        left = struct.pack('>Q', 0)
        uploaded = struct.pack('>Q', 0)

        event = struct.pack('>I', 0)
        ip = struct.pack('>I', 0)
        key = struct.pack('>I', 0)
        num_want = struct.pack('>i', -1)
        port = struct.pack('>h', 8000)

        msg = (conn_id + action + trans_id + info_hash + peer_id + downloaded +
                left + uploaded + event + ip + key + num_want + port)

        return msg, trans_id, action

    def send_msg(self,conn, sock, msg, trans_id, action, size):
        sock.sendto(msg, conn)
        try:
            response = sock.recv(2048)
        except socket.timeout as err:
            logging.debug(err)
            return
            #logging.debug("Connecting again...")
            #return self.send_msg(conn, sock, msg, trans_id, action, size)
        if len(response) < size:
            logging.debug("Did not get full message. Connecting again...")
            return self.send_msg(conn, sock, msg, trans_id, action, size)

        if action != response[0:4] or trans_id != response[4:8]:
            logging.debug("Transaction or Action ID did not match. Trying again...")
            return self.send_msg(conn, sock, msg, trans_id, action, size)

        return response

    def scrape_udp(self,torrent, announce):
        try:
            parsed = urlparse(announce)
            ip = socket.gethostbyname(parsed.hostname)

            if ip == '127.0.0.1':
                return False
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(3)
            conn = (ip, parsed.port)
            msg, trans_id, action = self.make_connection_id_request()
            response = self.send_msg(conn, sock, msg, trans_id, action, 16)
            if response == None:
                return ""

            conn_id = response[8:]
            msg, trans_id, action = self.make_announce_input(torrent.info_hash, conn_id, torrent.peer_id)
            response = self.send_msg(conn, sock, msg, trans_id, action, 20)
            if response == None or response == "":
                return ""
            peersByte = response[20:]
            raw_bytes = [ord(c) for c in peersByte]
            for i in range(len(raw_bytes) / 6):
                start = i * 6
                end = start + 6
                ip = ".".join(str(i) for i in raw_bytes[start:end - 2])
                port = raw_bytes[end - 2:end]
                port = port[1] + port[0] * 256
                self.newpeersQueue.put([ip, port])
        except:
            pass



import requests
from pubsub import pub
from collections import defaultdict


class HttpPeer(object):
    def __init__(self, torrent, url):
        self.readBuffer = b""
        self.torrent = torrent
        self.handshake(url)

    def handshake(self, url):
        if not url:
            raise Exception
        resp = requests.head(url)
        if resp.headers.get('Accept-Ranges', False) and resp.headers.get('ETag', False):  # if it was a full-url
            self.url = '/'.join(url.split('/')[0:-1]) + '/'
            self.etag = resp.headers.get('etag', None)
        elif url[-1] == '/':  # maybe we need to construct a full-url
            directory = self.torrent.torrentFile.get('info', {}).get('name')
            some_filename = self.torrent.torrentFile.get('info', {}).get('files', [{'path': ['']}])[0].get('path')[0]
            compound_url = url + directory + '/' + some_filename
            resp = requests.head(compound_url)
            if resp.headers.get('Accept-Ranges', False) and resp.headers.get('ETag', False):  # if it wasn't a full-url
                self.url = url + directory + '/'
                self.etag = resp.headers.get('etag', None)
            else:
                raise Exception

    def get_pieces(self, piecesManager):
        # TODO: Add another way to exit this loop when we don't hit the max size
        size = 0
        temp_size = 0
        max_size_to_download = 1500000
        temp_pieces = []
        pieces = []
        for idx, b in enumerate(piecesManager.bitfield):
            piece = piecesManager.pieces[idx]
            if not b and not piece.finished:
                temp_pieces.append(piece)
                temp_size += piece.pieceSize

            if size < temp_size:
                size = temp_size
                pieces = temp_pieces

            if size > max_size_to_download or len(pieces) > 5:
                return pieces

            if b:
                temp_pieces = []
                temp_size = 0
        return pieces

    def request_ranges(self, pieces_by_file):
        responses = {}
        for filename, pieces in pieces_by_file.items():
            start = pieces[0].get_file_offset(filename)
            directory = self.torrent.torrentFile.get('name')
            end = start
            for piece in pieces:
                end += piece.get_file_length(filename)
            resp = requests.get(self.url + filename, headers={'Range': 'bytes=' + str(start) + '-' + str(end)})
            responses[filename] = (resp, start)
        return responses

    def publish_responses(self, responses, pieces_by_file):
        unique_pieces = set()
        for piece_list in pieces_by_file.values():
            for piece in piece_list:
                unique_pieces.add(piece)

        for piece in unique_pieces:
            if piece.finished:
                continue
            for f in piece.files:
                filename = f.get('path').split('/')[-1]
                resp = responses[filename][0]
                resp_start = responses[filename][1]
                length = piece.get_file_length(filename)
                pieceOffset = piece.get_piece_offset(filename)
                fileOffset = piece.get_file_offset(filename) - resp_start
                piece.pieceData += resp.content[fileOffset: fileOffset + length]

            blockOffset = 0
            try:
                for idx in range(int(len(piece.pieceData)/piece.BLOCK_SIZE)):
                    block_size = piece.blocks[idx][1]
                    pub.sendMessage('PiecesManager.Piece', piece=(piece.pieceIndex, blockOffset, piece.pieceData[blockOffset: blockOffset + block_size]))
                    blockOffset += block_size
            except Exception:
                pass

    def construct_pieces_by_file(self, pieces):
        pieces_by_file = defaultdict(list)
        for piece in pieces:
            for f in piece.files:
                filename = f.get('path').split('/')[-1]
                pieces_by_file[filename].append(piece)
        return pieces_by_file



import time
import logging
import json
import io
import os
import bencode
import requests
import utils
import pkg_resources
from .Client import Client
from .Torrent import Torrent


def get(hash, datastore=None, name=None):
    torrent_dir = utils.get_torrent_dir(datastore=datastore, name=name)
    torrent = Torrent(hash, torrent_dir)
    timestamp = utils.read_timestamp(hash)
    if utils.timestamp_is_recent(timestamp) and utils.filenames_present(torrent, datastore):
        return torrent_dir + torrent.torrentFile['info']['name']
    else:
        return Client(hash, torrent_dir).start()



from setuptools import setup, find_packages
import academictorrents as at

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='academictorrents',
    version=at.__version__,
    description='Academic Torrents Python APIs',
    long_description=readme,
    author='Martin Weiss, Alexis Gallepe, Jonathan Nogueira',
    author_email='contact@academictorrents.com',
    url='https://github.com/AcademicTorrents/python-r-api',
    classifiers=(
        "Programming Language :: Python :: 2.6"
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ),
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'bencode.py==2.0.0',
        'bitstring==3.1.5',
        'PyPubSub==3.3.0',
        'requests==2.7.0',
        'future==0.16.0'
    ]
)



__author__ = 'alexisgallepe'

from pubsub import pub

class RarestPieces(object):
    def __init__(self, piecesManager):

        self.piecesManager = piecesManager
        self.rarestPieces = []

        for pieceNumber in range(self.piecesManager.numberOfPieces):
            peersByPieceIndex = {"idPiece":pieceNumber, "numberOfPeers":0, "peers":[]}
            self.rarestPieces.append(peersByPieceIndex)

        #pub.subscribe(self.peersBitfield, 'RarestPiece.updatePeersBitfield')


    def peersBitfield(self,bitfield=None,peer=None,pieceIndex=None):

        if len(self.rarestPieces) == 0:
            raise("no more piece")

        # Piece complete
        try:
            if not pieceIndex == None:
                self.rarestPieces.__delitem__(pieceIndex)
        except:
            pass

        # Peer's bitfield updated
        else:
            for i in range(len(self.rarestPieces)):
                if bitfield[i] == 1 and peer not in self.rarestPieces[i]["peers"]:
                    self.rarestPieces[i]["peers"].append(peer)
                    self.rarestPieces[i]["numberOfPeers"] = len(self.rarestPieces[i]["peers"])

    def getSortedPieces(self):
        return sorted(self.rarestPieces, key=lambda x:x['numberOfPeers'])


__author__ = 'alexisgallepe'

import time
from . import Peer
from threading import Thread
from pubsub import pub


class PeerSeeker(Thread):
    def __init__(self, newpeersQueue, torrent):
        Thread.__init__(self)
        self.stopRequested = False
        self.newpeersQueue = newpeersQueue
        self.torrent = torrent
        self.peerFailed = [("", "")]
        self.setDaemon(True)

    def requestStop(self):
        self.stopRequested = True

    def run(self):
        while not self.stopRequested:
            peer = self.newpeersQueue.get()
            if not (peer[0], peer[1]) in self.peerFailed:
                p = Peer.Peer(self.torrent, peer[0], peer[1])
                if not p.connectToPeer(3):
                    self.peerFailed.append((peer[0], peer[1]))
                else:
                    print("add new peer: " + p.ip)
                    pub.sendMessage('PeersManager.newPeer', peer=p)
